{% extends "siteconfig/settings.html" %}
{% load djblets_utils %}


{% block after_field_sets %}
{%  if form.Meta.subforms %}
{%   for subform_info in form.Meta.subforms %}
{%    with subforms=form|getattr:subform_info.subforms_attr %}
{%     if subforms %}
{%      if subform_info.controller_field %}
{%       with subform_controller_field=form|getitem:subform_info.controller_field %}
{%        include "forms/subforms.html" with subform_controller=subform_controller_field.id_for_label %}
{%       endwith %}
{%      else %}
{%       include "forms/subforms.html" %}
{%      endif %}
{%     endif %}
{%    endwith %}
{%   endfor %}
{%  endif %}

{{block.super}}
{% endblock after_field_sets %}
