/** Base styles used throughout the Admin UI. */

@import (reference) "rb/css/defs.less";
@import (reference) "rb/css/mixins/theme.less";


#rb-ns-ui() {
  .page-colors() {
    .admin() {
      @topbar-bg: #rb-ns-ui.colors[@yellow-70];
    }
  }
}


.admin-page {
  --ink-p-header-bg:
    var(--if-light, var(--ink-p-yellow-300))
    var(--if-dark, var(--ink-p-cool-grey-300));
}
