{% load actions djblets_utils gravatars i18n rb_extensions %}

<div class="rb-c-topbar__header" id="headerbar">
{% include "base/branding.html" %}
 <div class="rb-c-topbar__nav-toggle" id="nav_toggle" role="button" aria-label="{% trans 'Toggle navigation' %}"><span class="ink-i-mobile-menu"></span></div>

 <div class="rb-c-topbar__search" roel="search">
{% if siteconfig_settings.search_enable %}
  <form method="get" action="{% url 'search' %}">
{% endif %}
 <div class="rb-c-search-field" id="search">
   <span class="fa fa-search"></span>
   <input class="rb-c-search-field__input" type="search" name="q" placeholder="{% if siteconfig_settings.search_enable %}{% trans "Search" %}{% else %}{% trans "Jump to..." %}{% endif %}" id="search_field" {% attr "value" %}{{query}}{% endattr %} />
 </div>
{% if siteconfig_settings.search_enable %}
  </form>
{% endif %}
 </div>

 <nav class="rb-c-topbar__actions rb-c-actions" aria-label="{% trans 'Account and product' %}" role="presentation">
  <menu class="rb-c-actions__content" id="accountnav" role="menu">
{% actions_html "header" %}
  </menu>
 </nav>
</div>
