{% extends "base.html" %}

{% load djblets_deco i18n %}

{% block title %}{% blocktrans %}{{PRODUCT_NAME}} is in read-only mode{% endblocktrans %}{% endblock %}

{% block content %}
{%  box 'important' %}
<h1>
{%   blocktrans %}
 {{PRODUCT_NAME}} is in read-only mode.
{%   endblocktrans %}
</h1>
<p>
{%   blocktrans %}
 The site is currently in read-only mode and some features are unavailable.
{%   endblocktrans %}
</p>
{%  endbox %}
{% endblock %}
