#custom_policy_editor {
  margin: 0;
  width: 65em;
}
#custom_policy_editor p {
  font-size: 110%;
  margin: 1em;
}
#custom_policy_editor .CodeMirror {
  border-left: 0;
  border-right: 0;
}
#page_authentication {
  /**
   * The API tokens configuration.
   *
   * Structure:
   *     <div class="rb-c-config-api-tokens">
   *      <div class="rb-c-config-api-tokens__main">
   *       <div class="rb-c-config-api-tokens__value">
   *        <input readonly="readonly" value="...">
   *       </div>
   *       <span class="fa fa-clipboard js-copy-token" title="Copy to clipboard"></span>
   *      </div>
   *      <div class="rb-c-config-api-tokens__info">
   *       <p class="rb-c-config-api-tokens__deprecation-notice">...</p>
   *       <p class="rb-c-config-api-tokens__token-state">...</p>
   *       <p class="rb-c-config-api-tokens__usage">...</p>
   *      </div>
   *      <div class="rb-c-config-api-tokens__actions"></div>
   *      <span class="rb-c-config-api-tokens__note"></span>
   *     </div>
   */
}
#page_authentication .rb-c-config-api-tokens {
  /**
     * Displays a deprecation notice for token.
     *
     * If the token is deprecated, this will display a deprecation notice. If
     * it is not deprecated this will display nothing.
     *
     * Structure:
     *     <p class="rb-c-config-api-tokens__deprecation-notice">...</p>
     */
  /**
     * Displays the state of the token.
     *
     * This will display whether the token expires, is expired, or is invalid.
     *
     * Modifiers:
     *     -has-expires:
     *         The token is valid and has a future expiration date. The
     *         expiration date will be shown.
     *
     *     -is-expired:
     *         The token is expired. The date that the token expired on
     *         will be shown.
     *
     *     -is-invalid:
     *         The token is invalid. The date that the token was invalidated
     *         on will be shown.
     *
     * Structure:
     *     <p class="rb-c-config-api-tokens__token-state [modifiers]">...</p>
     */
  /**
     * Displays information on how the token has been used.
     *
     * Modifiers:
     *     -has-last-used:
     *         The token is valid and has been used before. The date that
     *         the token was last used will be shown.
     *
     * Structure:
     *     <p class="rb-c-config-api-tokens__usage [modifiers]">...</p>
     */
}
#page_authentication .rb-c-config-api-tokens__actions {
  white-space: nowrap;
}
#page_authentication .rb-c-config-api-tokens__deprecation-notice:before {
  content: "\f071";
  font-family: 'FontAwesome';
}
#page_authentication .rb-c-config-api-tokens__info {
  line-height: 0;
}
#page_authentication .rb-c-config-api-tokens__main {
  align-items: center;
  display: flex;
  white-space: normal;
}
#page_authentication .rb-c-config-api-tokens__note {
  color: #888;
}
#page_authentication .rb-c-config-api-tokens__note.empty {
  color: #CCC;
  font-style: italic;
}
#page_authentication .rb-c-config-api-tokens__token-state.-is-expired,
#page_authentication .rb-c-config-api-tokens__token-state.-is-invalid {
  color: red;
}
#page_authentication .rb-c-config-api-tokens__token-state input.save {
  margin-left: 0.5em;
}
#page_authentication .rb-c-config-api-tokens__value {
  font-family: var(--ink-ff-monospace);
  font-size: 120%;
  flex-grow: 1;
  margin-right: 1em;
}
#page_authentication .rb-c-config-api-tokens__value input {
  font-family: var(--ink-ff-monospace);
  text-overflow: ellipsis;
  width: 100%;
}
#page_authentication .rb-c-config-api-tokens__value input:not(:hover) {
  border-color: transparent;
}
#page_authentication .rb-c-config-api-tokens .js-copy-token {
  font-size: 120%;
  cursor: pointer;
}
#page_authentication .rb-c-config-api-tokens .config-forms-list-action-label-policy-custom {
  color: var(--ink-p-link-fg);
  cursor: pointer;
  text-decoration: underline;
}
#page_authentication .rb-c-config-api-tokens .config-forms-list-action-policy {
  text-align: right;
}
#page_authentication .rb-c-config-api-tokens .rb-icon {
  vertical-align: middle;
}
#page_groups .config-group-display-name {
  color: #A0A0A0;
  font-size: 90%;
}
#page_groups .config-group-name {
  display: inline-block;
  min-width: 15em;
  padding-right: 2em;
}
#page_groups .config-forms-list-action-join,
#page_groups .config-forms-list-action-leave {
  min-width: 3em;
}
#page_oauth2 .app-entry-wrapper {
  overflow: hidden;
}
#page_oauth2 .config-app-name.disabled a {
  font-style: italic;
}
#page_oauth2 .config-app-name .rb-icon {
  vertical-align: middle;
}
#page_oauth2 .config-app-name a {
  vertical-align: middle;
}
#page_oauth2 .disabled-warning {
  margin-bottom: 1em;
}
#page_oauth2 p {
  line-height: normal;
}
.oauth-add-app {
  margin-top: 1em;
}
.edit-oauth2-app #id_redirect_uris_container li {
  padding: 0.25em 0;
}
.edit-oauth2-app .box-main {
  padding: 0;
}
.edit-oauth2-app .list-edit-widget ul {
  list-style: none;
}
.edit-oauth2-app .regenerate-secret-button {
  margin-top: 1px;
}
/*# sourceMappingURL=pages/my-account.css.4bd05b553767.map */