{% extends "datagrids/listview.html" %}
{% load djblets_utils i18n reviewtags %}

{% block datagrid_titlebox %}
{%  if datagrid.group %}{% star datagrid.group %}{% endif %}
{{block.super}}
{% endblock datagrid_titlebox %}

{% block datagrid_filters %}
<li class="datagrid-showclosed">
{%  if datagrid.show_closed %}
 <a href="{% querystring 'update' 'show-closed=0' 'page=1' %}" rel="nofollow">{% trans "Hide closed" %}</a>
{%  else %}
 <a href="{% querystring 'update' 'show-closed=1' 'page=1' %}" rel="nofollow">{% trans "Show closed" %}</a>
{%  endif %}
</li>
{% endblock datagrid_filters%}
