/* Note: These are roughly in dependency order. */
@import 'rb/css/pages/admin/base.less';
@import 'rb/css/pages/admin/change-form.less';
@import 'rb/css/pages/admin/change-list.less';
@import 'rb/css/pages/admin/auth.less';
@import 'rb/css/pages/admin/dashboard.less';
@import 'rb/css/pages/admin/database.less';
@import 'rb/css/pages/admin/filediffs.less';
@import 'rb/css/pages/admin/integrations.less';
@import 'rb/css/pages/admin/log-viewer.less';
@import 'rb/css/pages/admin/repositories.less';
@import 'rb/css/pages/admin/security-center.less';
@import 'rb/css/pages/admin/webhooks.less';
@import 'rb/css/pages/admin/widgets.less';
@import 'rb/css/ui/admin/filters.less';
@import 'rb/css/ui/admin/license.less';
@import 'rb/css/ui/admin/search.less';
@import 'rb/css/ui/admin/widgets.less';
