{% extends "notifications/email_base.html" %}
{% load djblets_utils %}
{% load markdown_email rb_extensions %}

{% block content %}
{% include "notifications/review_email_content.html" %}
{% endblock %}

<p>- {{review.user|user_displayname}}</p>

{% block footer %}
<br />
<p>On {{review_request.time_emailed|date:"F jS, Y, P T"}}, {{review_request.submitter|user_displayname}} wrote:</p>

{%  include "notifications/review_request_email_content.html" %}
{% endblock %}
