{% extends "admin/change_form.html" %}

{% load djblets_forms i18n %}


{% block page_title %}{{form.Meta.title}}{% endblock page_title %}
{% block form_title %}{{form.Meta.title}}{% endblock form_title %}


{% block field_sets %}
{%  include "djblets_forms/admin/form_fieldsets.html" %}
{% endblock field_sets %}


{% block submit_row_buttons %}
{%  with show_save=True %}
{{block.super}}
{%  endwith %}
{% endblock submit_row_buttons %}
