{% extends "actions/action_base.html" %}

{% block action_content %}
<a {{action_attrs}}
   href="{{url}}"
   role="menuitem"
   data-custom-rendered="true">
 <h4>
{% if action.icon_class %}
  <span class="{{action.icon_class}}"></span>
{% endif %}
  {{verbose_label|default:label}}
 </h4>
{%  for description_part in action.description %}
 <p>
  {{description_part}}
 </p>
{%  endfor %}
</a>
{% endblock action_content %}
