{% if field.field.is_hidden %}
{{field.render}}
{% else %}
<div class="{{field.classes}}">
{%  if field.show_errors %}
 <div class="rb-c-form-field__errors">
  {{field.errors}}
 </div>
{%  endif %}
{%  if field.has_field_first %}
 <div class="rb-c-form-field__input">
  {{field.render}}
  {{field.label_tag}}
 </div>
{%   if field.field.help_text %}
  <div class="rb-c-form-field__help">
   <span>{{field.field.help_text|safe}}</span>
  </div>
{%   endif %}
{%  else %}
 {{field.label_tag}}
 <div class="rb-c-form-field__input">
  {{field.render}}
{%   if field.field.help_text %}
  <div class="rb-c-form-field__help">
   <span>{{field.field.help_text|safe}}</span>
  </div>
{%   endif %}
 </div>
{%  endif %}
</div>
{% endif %}
