{% extends "actions/group_action.html" %}
{% load djblets_utils %}


{% block action_container_tag %}li{% endblock %}
{% block action_container_css_class %}rb-c-sidebar__section{% endblock %}
{% block action_has_container %}true{% endblock %}


{% block action_container_attrs %}
{{block.super}}
{{action_attrs}}
 aria-labelledby="{{action.action_id}}__label"
{% endblock action_container_attrs %}


{% block action_content %}
{%  definevar "sidebar_items" %}{{block.super}}{% enddefinevar %}
<header class="rb-c-sidebar__section-header"
        id="{{action.action_id}}__label">{{label}}</header>
<ul class="rb-c-sidebar__items">
{%  block action_sidebar_items %}{{sidebar_items}}{% endblock %}
</ul>
{% endblock action_content %}
