@import (reference) "rb/css/layout/helpers.less";


#search_results {
  margin: var(--ink-u-spacing-l);

  .paginator {
    border-top: var(--ink-g-border-container);
    margin-top: var(--ink-u-spacing-l);
    padding-top: var(--ink-u-spacing-l);
    font-size: var(--ink-u-font-ml);

    a {
      text-decoration: none;
    }

    a, b {
      margin-right: 0.5em;
    }
  }

  .search-results-filters {
    border-bottom: var(--ink-g-border-container);
    margin-bottom: var(--ink-u-spacing-l);

    a {
      display: inline-block;
      font-size: var(--ink-u-font-l);
      text-decoration: none;
      padding: 0.5em 1em;
      border-bottom: var(--ink-u-border-thick) transparent solid;

      &.active {
	      border-bottom: var(--ink-u-border-thick)
	                     var(--ink-p-container-border-color)
		                   solid;
      }
    }
  }

  .search-results-stats {
    font-size: var(--ink-u-font-ml);
  }

  .search-result {
    display: flex;
    flex-direction: column;
    gap: var(--ink-u-spacing-xs);
    line-height: 1.5;
    margin-bottom: var(--ink-u-spacing-l);

    &.-has-icon {
      flex-direction: row;
      gap: var(--ink-u-spacing-m);
    }

    a {
      text-decoration: none;
    }

    h2 {
      font-size: var(--ink-u-font-l);
      font-weight: normal;
      margin: 0;
    }

    .search-result-by {
      color: var(--ink-p-fg-weak);
    }
  }

  form {
    margin-top: var(--ink-u-spacing-l);
    max-width: 60em;
  }
}
