import { Action } from './models/actionModel';
import { GroupAction } from './models/groupActionModel';
import { MenuAction } from './models/menuActionModel';
import { ActionView } from './views/actionView';
import { ButtonActionView } from './views/buttonActionView';
import { MenuActionView, MenuItemActionView } from './views/menuActionView';


export const Actions = {
    Action,
    ActionView,
    ButtonActionView,
    GroupAction,
    MenuAction,
    MenuActionView,
    MenuItemActionView,
};
