<li class="rb-c-sidebar__nav-item{% if active %} -is-active{% endif %}{% if css_classes %} {{css_classes|join:' '}}{% endif %}">
{% if icon_name %}
 <div class="rb-c-sidebar__icon rb-icon {{icon_name|default:''}}"></div>
{% endif %}
 <a href="{{url}}" class="rb-c-sidebar__item-label">{{label}}</a>
{% if has_count %}
 <span class="rb-c-sidebar__item-value {% if count == 0 %}-is-dimmed{% endif %}">{{count}}</span>
{% endif %}
 </a>
</li>
