/**
 * Styles and variables for the admin search component.
 */


/**
 * The search field shown on change list pages.
 *
 * Structure:
 *     <form class="rb-c-admin-search" method="GET">
 *      <div class="rb-c-search-field">...</div>
 *      <div class="rb-c-sadmin-search__result-summary">...</div>
 *     </form>
 */
.rb-c-admin-search {
  margin-bottom: 1em;

  /**
   * The summary of search results.
   *
   * Structure:
   *     <div class="rb-c-admin-search__result-summary">...</div>
   */
  &__result-summary {
    margin: 1em 0;
  }
}
