<td class="{{css_class}}"{% if column_state.last %} colspan="2"{% endif %} onclick="javascript:window.location = '{{url|escapejs}}'; return false;">
<a href="{{url}}" class="{{link_css_class}}" data-review-request-url="{{review_request_url}}">{{data}}</a>
</td>
