{% load actions reviewtags %}

<header class="rb-c-review-request__header">
 <div class="rb-c-actions" role="presentation">
  <menu class="rb-c-actions__content -is-left" role="menu">
{% actions_html "review-request-left" %}
  </menu>
  <menu class="rb-c-actions__content -is-right -has-mobile-menu"
        role="menu">
   <li class="rb-c-actions__action rb-o-mobile-menu-label" role="presentation">
    <a href="#"
       aria-controls="review-request-mobile-actions-menu-content"
       aria-expanded="false"
       aria-haspopup="true"
       ><span class="ink-i-mobile-menu" aria-hidden="true"></span></a>
   </li>
   <div class="rb-o-mobile-menu" id="review-request-mobile-actions-menu-content">
{% actions_html "review-request" %}
   </div>
  </menu>
 </div>

 <menu class="rb-c-review-request-tabs" role="menu">
{% for tab in tabs %}
  <li class="rb-c-review-request-tabs__tab{% if tab.active or tab.url == request.path %} -is-active{% endif %}"
      role="presentation">
   <a role="menuitem"
      href="{{tab.url|add_view_draft_query:viewing_user_draft}}"
      {% if tab.active or tab.url == request.path %}aria-current="page"{% endif %}
      >{{tab.text}}</a>
  </li>
{% endfor %}
 </menu>
</header>
