import '../admin/collections/tests';
import '../admin/models/tests';
import '../admin/views/tests';
import '../common/models/tests';
import '../common/resources/collections/tests';
import '../common/resources/models/tests';
import '../common/views/tests';
import '../reviews/collections/tests';
import '../reviews/models/tests';
import '../reviews/views/tests';
import '../ui/models/tests';
import '../ui/views/tests';
