{% load avatars i18n tz %}

<li class="rb-c-sidebar__section -is-desktop-only">
 <ul class="rb-c-sidebar__items">
  <li class="rb-c-sidebar__item" id="user_page_profile">
{% if siteconfig_settings.avatars_enabled %}
   {% avatar profile_user 175 %}
{% endif %}

   <h2 class="nickname">{{profile_user.username}}</h2>

{% if show_profile %}
   <div class="fn">
   {{profile_user.first_name}} {{profile_user.last_name}}
   </div>

{%  if request.user.is_authenticated %}
   <div>
    <a class="email" href="mailto:{{profile_user.email}}">{{profile_user.email}}</a>
   </div>
{%  endif %}

{%  localtime on %}
   <div class="logged-in">
{%   blocktrans with profile_user.last_login|localtime|date as last_login_date %}
    Last logged in {{last_login_date}}
{%   endblocktrans %}
   </div>

   <div class="joined">
{%   blocktrans with profile_user.date_joined|localtime|date as joined_date %}
    Joined {{joined_date}}
{%   endblocktrans %}
   </div>
{%  endlocaltime %}
{% endif %}{# show_profile #}
  </li>
 </ul>
</li>
