/**
 * Common styles for avatars.
 */


/**
 * A standard avatar on the page.
 *
 * This will style the avatar to be shown as a circle, and help
 * vertically-align it to make it appear correctly next to text.
 */
.djblets-o-avatar {
  border-radius: 50%;
  vertical-align: middle;
}
