{% load djblets_utils static %}

<div class="rb-c-admin-widget{% if widget.css_classes %} {{widget.css_classes}}{% endif %}" id="{{widget.dom_id}}">
 <header class="rb-c-admin-widget__header">
  <h1>{{widget.name}}</h1>
  <ul class="rb-c-admin-widget__actions">
{% block widget_header_actions %}{% endblock %}
  </ul>
 </header>
 <div class="rb-c-admin-widget__content">
{% block widget_content %}{% endblock %}
 </div>
 <footer class="rb-c-admin-widget__footer">
  <ul class="rb-c-admin-widget__actions">
{% block widget_footer_actions %}{% endblock %}
  </ul>
 </footer>
</div>
