import './baseCommentModelTests';
import './baseResourceModelTests';
import './defaultReviewerModelTests';
import './draftReviewModelTests';
import './fileAttachmentModelTests';
import './reviewGroupModelTests';
import './reviewModelTests';
import './reviewReplyModelTests';
import './reviewRequestModelTests';
