{% extends "base.html" %}

{% block css %}
{{block.super}}
<style type="text/css">
.icon-row {
  margin: 1em;
}
.rb-icon {
  border: 1px solid black;
}
</style>
{% endblock css %}

{% block title %}Icons{% endblock %}

{% block main_content %}
<div class="icon-row"><span class="rb-icon rb-icon-issue-dropped"></span> issue-dropped</div>
<div class="icon-row"><span class="rb-icon rb-icon-issue-open"></span> issue-open</div>
<div class="icon-row"><span class="rb-icon rb-icon-issue-resolved"></span> issue-resolved</div>
<div class="icon-row"><span class="rb-icon rb-icon-collapse-review"></span> collapse-review</div>
<div class="icon-row"><span class="rb-icon rb-icon-expand-review"></span> expand-review</div>
<div class="icon-row"><span class="rb-icon rb-icon-collapse"></span> collapse</div>
<div class="icon-row"><span class="rb-icon rb-icon-expand"></span> expand</div>
<div class="icon-row"><span class="rb-icon rb-icon-dropdown-arrow"></span> dropdown-arrow</div>
<div class="icon-row"><span class="rb-icon rb-icon-datagrid-comment-draft"></span> datagrid-comment-draft</div>
<div class="icon-row"><span class="rb-icon rb-icon-datagrid-comment-shipit"></span> datagrid-comment-shipit</div>
<div class="icon-row"><span class="rb-icon rb-icon-datagrid-comment"></span> datagrid-comment</div>
<div class="icon-row"><span class="rb-icon rb-icon-datagrid-shipit"></span> datagrid-shipit</div>
<div class="icon-row"><span class="rb-icon rb-icon-new-updates"></span> new-updates</div>
<div class="icon-row"><span class="rb-icon rb-icon-edit"></span> edit</div>
<div class="icon-row"><span class="rb-icon rb-icon-delete"></span> delete</div>
<div class="icon-row"><span class="rb-icon rb-icon-diff-collapse-chunk"></span> diff-collapse-chunk</div>
<div class="icon-row"><span class="rb-icon rb-icon-diff-expand-above"></span> diff-expand-above</div>
<div class="icon-row"><span class="rb-icon rb-icon-diff-expand-all"></span> diff-expand-all</div>
<div class="icon-row"><span class="rb-icon rb-icon-diff-expand-below"></span> diff-expand-below</div>
<div class="icon-row"><span class="rb-icon rb-icon-diff-expand-header"></span> diff-expand-header</div>
<div class="icon-row"><span class="rb-icon rb-icon-admin-enabled"></span> admin-enabled</div>
<div class="icon-row"><span class="rb-icon rb-icon-admin-disabled"></span> admin-disabled</div>
<div class="icon-row"><span class="rb-icon rb-icon-admin-add"></span> admin-add</div>
<div class="icon-row"><span class="rb-icon rb-icon-star-off"></span> star-off</div>
<div class="icon-row"><span class="rb-icon rb-icon-star-on"></span> star-on</div>
<div class="icon-row"><span class="rb-icon rb-icon-warning"></span> warning</div>
<div class="icon-row"><span class="rb-icon rb-icon-search"></span> search</div>
<div class="icon-row"><span class="rb-icon rb-icon-search-dark"></span> search-dark</div>
<div class="icon-row"><span class="rb-icon rb-icon-admin-collapse"></span> admin-collapse</div>
<div class="icon-row"><span class="rb-icon rb-icon-admin-expand"></span> admin-expand</div>
<div class="icon-row"><span class="rb-icon rb-icon-jump-to"></span> jump-to</div>
<div class="icon-row"><span class="rb-icon rb-icon-range-slider"></span> range-slider</div>
<div class="icon-row"><span class="rb-icon rb-icon-download"></span> download</div>
<div class="icon-row"><span class="rb-icon rb-icon-remove-widget"></span> remove-widget</div>
<div class="icon-row"><span class="rb-icon rb-icon-archive-off"></span> archive-off</div>
<div class="icon-row"><span class="rb-icon rb-icon-archive-on"></span> archive-on</div>
<div class="icon-row"><span class="rb-icon rb-icon-datagrid-open-issues"></span> datagrid-open-issues</div>
<div class="icon-row"><span class="rb-icon rb-icon-datagrid-column-shipits-issues"></span> datagrid-column-shipits-issues</div>
<div class="icon-row"><span class="rb-icon rb-icon-shipit"></span> shipit</div>
<div class="icon-row"><span class="rb-icon rb-icon-issue-verifying"></span> issue-verifying</div>
<div class="icon-row"><span class="rb-icon rb-icon-datagrid-issue-verifying"></span> datagrid-issue-verifying</div>
<div class="icon-row"><span class="rb-icon rb-icon-compose-review"></span> compose-review</div>
<div class="icon-row"><span class="rb-icon rb-icon-edit-review"></span> edit-review</div>
<div class="icon-row"><span class="rb-icon rb-icon-create-review"></span> create-review</div>
<div class="icon-row"><span class="rb-icon rb-icon-review-num-files"></span> review-num-files</div>
{% endblock main_content %}
