{% load djblets_utils %}
<html>
 <body style="font-family: Verdana, Arial, Helvetica, Sans-Serif;">
  <table bgcolor="#f9f3c9" width="100%" cellpadding="8" style="border: 1px #c9c399 solid;">
   <tr>
    <td>
     This is an automatically generated e-mail.
    </td>
   </tr>
  </table>

  <p>Hi {{user|user_displayname}},</p>
  <p>
   A new API token has been added to your {{PRODUCT_NAME}} account on
   <a href="{{site_root_url}}">{{site_root_url}}</a>.
  </p>
  <p>
   The API token ID starts with <code>{{partial_token}}</code> and was added
   {{api_token.time_added|date:"F jS, Y, P T"}}.
  </p>
{% if client %}
  <p>
   This token was automatically created for {{client}}. If you did not just
   authenticate to Review Board for {{client}}, you should revoke this
   token on your <a href="{{api_token_url}}">API Tokens</a> page,
   change your password, and talk to your administrator.
  </p>
{% else %}
  <p>
   If you did not create this token, you should revoke it on your
   <a href="{{api_token_url}}">API Tokens</a>
   page, change your password, and talk to your administrator.
  </p>
{% endif %}
 </body>
</html>
