import {
    ReviewReplyDraftBannerView,
    ReviewReplyDraftStaticBannerView,
} from './views/reviewReplyDraftBannerView';


/* Define a namespace for RB.ReviewRequestPage. */
export const ReviewRequestPage = {
    ReviewReplyDraftBannerView,
    ReviewReplyDraftStaticBannerView,
};
