{% extends "actions/action_base.html" %}
{% load actions %}


{% block action_container_role %}group{% endblock %}


{% block action_content %}
{%  child_actions_html %}
{% endblock action_content %}
