@import (reference) "rb/css/defs.less";

.upload-diff-ui() {
  #processing-diff,
  #uploading-diffs {
    .spinner {
      width: 16px;
      height: 16px;
      font-size: 16px;
      margin: auto;
    }
  }

  #error-contents,
  #parent-diff-error-contents {
    line-height: 12pt;
    margin-bottom: 1em;
    position: relative;

    img {
      margin-right: 5px;
      vertical-align: sub;
    }
  }

  .input {
    border-radius: @box-border-radius;
    position: relative;
    text-align: center;
    vertical-align: middle;

    > form {
      display: flex;
      align-items: center;
      justify-content: center;
      gap: 1ch;
    }

    .startover {
      position: absolute;
      bottom: 0;
      right: 5px;
      line-height: 14pt;
    }
  }

  .drag-hover {
    box-shadow: inset @box-shadow;
  }
}
