"use strict";

/**
 * A collection of CommitsHistoryDiffEntries.
 */
RB.CommitHistoryDiffEntryCollection = Backbone.Collection.extend({
  model: RB.CommitHistoryDiffEntry
});

//# sourceMappingURL=commitHistoryDiffEntryCollection.js.map