/**
 * Base support for building extension hook points.
 *
 * This is a re-export of the base extension hook point infrastructure found in
 * Djblets.
 *
 * Version Added:
 *     8.0
 */

export { ExtensionHookPoint } from 'djblets/extensions';
