{% extends "datagrid/listview.html" %}

{% block datagrid_titlebox %}
<header class="rb-c-content-header rb-c-datagrid-header -is-main{% if datagrid.tabs %} -has-tabs{% endif %}">
{%  if datagrid.tabs %}
 <ul class="datagrid-title-tabs rb-c-tabs">
{%   for tab_title, url in datagrid.tabs %}
  <li class="rb-c-tabs__tab{% if tab_title == datagrid.tab_title %} -is-active{% endif %}">
   <a class="rb-c-tabs__tab-label" href="{{url}}">{{tab_title}}</a>
  </li>
{%   endfor %}
 </ul>
{%  else %}
 <h1 class="rb-c-content-header__title">{{datagrid.title}}</h1>
{%  endif %}

{%  block datagrid_filters_container %}
 <div class="rb-c-content-header__actions datagrid-top-filters">
  <div class="rb-c-content-header__action datagrid-filters-menu">
   <span class="fa fa-cog"></span>
  </div>
  <ul class="rb-c-content-header__action datagrid-filters">
{%   block datagrid_filters %}{% endblock %}
  </ul>
 </div>
{%  endblock datagrid_filters_container %}
</header>
{% endblock datagrid_titlebox %}
