{% extends "base.html" %}
{% load djblets_js i18n pipeline %}

{% block title %}{% trans "New Review Request" %}{% endblock %}

{% block extrahead %}
<meta name="viewport" content="width=device-width, initial-scale=1, maximum-scale=1" />
{% endblock %}

{% block page_class %}new-review-request-page -has-sidebar -is-content-flush-on-mobile -is-content-full-page{% endblock %}
{% block page-sidebar-css-classes %}-is-connected{% endblock %}

{% block css %}
{{block.super}}
{%  stylesheet 'newReviewRequest' %}
{% endblock css %}

{% block scripts-post %}
{{block.super}}
{%  javascript 'newReviewRequest' %}
{% endblock scripts-post %}

{% block js-page-view-type %}RB.NewReviewRequestView{% endblock %}
{% block js-page-model-type %}RB.NewReviewRequest{% endblock %}
{% block js-page-model-attrs %}{{page_model_attrs|json_dumps}}{% endblock %}
{% block js-page-model-options %}{parse: true}{% endblock %}
