"""Hosting service for Forgejo.

Version Added:
    8.0
"""

from __future__ import annotations

from reviewboard.hostingsvcs.forgejo.service import Forgejo


__all__ = [
    'Forgejo',
]

__autodoc_excludes__ = __all__
