import './commentDialogViewTests';
import './commentIssueBarViewTests';
import './diffReviewableViewTests';
import './diffViewerPageViewTests';
import './fileAttachmentThumbnailViewTests';
import './reviewDialogViewTests';
import './reviewRequestEditorViewTests';
import './reviewRequestFieldViewsTests';
import './reviewablePageViewTests';
import './textBasedReviewableViewTests';
