"""Integrations support for Review Board.

This module provides the Review Board functionality needed to create
integrations for third-party services. It builds upon Djblets's integrations
foundation, offering some additional utilities for more easily creating
manageable integrations.
"""
