/**
 * Base support for building extensions.
 *
 * This is a re-export of the base extension infrastructure found in Djblets.
 *
 * Version Added:
 *     8.0
 */

export {
    type BaseExtensionAttrs,
    Extension,
} from 'djblets/extensions';
