{% load reviewtags %}
<li class="rb-c-actions__action review-request-action-star -is-icon" role="presentation">
 <a href="#" id="{{dom_element_id}}" role="menuitem">{% star review_request %}</a>
</li>
