<li class="rb-c-sidebar__section{% if active %} -is-active{% endif %}{% if css_classes %} {{css_classes|join:' '}}{% endif %}">
 <header class="rb-c-sidebar__nav-section-header">
  <a href="{{url}}">
   {{label}}
  </a>
 </header>
</li>
