<fieldset class="rb-c-form-fieldset -is-subform -is-wide"
          data-subform-id="{{subform_id}}"
          {% if subform_controller %}data-subform-controller="{{subform_controller}}"{% endif %}
          {% if subform_group %}data-subform-group="{{subform_group}}"{% endif %}
          {% if subform_info.enable_checkbox %}data-subform-enabler="id_{{subform_id}}_enabled"{% endif %}
          disabled hidden>
 <legend class="rb-c-form-fieldset__name">{% spaceless %}
{%    if subform.Meta.title %}
  {{subform.Meta.title}}
{%    else %}
{{subform_id|title}}
{%    endif %}
{%   endspaceless %}</legend>
 <div class="rb-c-form-fieldset__content">
  <div class="rb-c-form-fieldset__fields">
{%   include "forms/fieldsets.html" with form=subform %}
  </div>
 </div>
</fieldset>
