{% extends "actions/action_base.html" %}


{% block action_content %}
<a {{action_attrs}}
   href="{{url}}"
   role="button">
{%  if action.icon_class %}
 <span class="{{action.icon_class}}"></span>
{%  endif %}
 {{label}}
</a>
{% endblock action_content %}
