{% load i18n %}

<div class="rb-c-form__actions-primary">
{% if show_save %}
 <button class="rb-c-form__action ink-c-button -is-primary" type="submit"
         name="_save">
  {% trans "Save" %}
 </button>
{% endif %}

{% if show_save_as_new %}
 <button class="rb-c-form__action ink-c-button"
         name="_saveasnew">
  {% trans "Save as new" %}
 </button>
{% endif %}

{% if show_save_and_add_another %}
 <button class="rb-c-form__action ink-c-button"
         name="_addanother">
  {% trans "Save and add another" %}
 </button>
{% endif %}

{% if show_save_and_continue %}
 <button class="rb-c-form__action ink-c-button"
         name="_continue">
  {% trans "Save and continue editing" %}
 </button>
{% endif %}
</div>

{% if show_delete_link %}
<div class="rb-c-form__actions-secondary">
 <a class="rb-c-form__action ink-c-button -is-danger" role="button"
    draggable="false" href="{{delete_url}}">{% trans "Delete" %}</a>
 </div>
{% endif %}
</div>
