/**
 * All view exports for Review Board's JavaScript extension support.
 *
 * Version Added:
 *     8.0
 */

export * from './baseCommentDialogHookView';
export * from './baseFileAttachmentThumbnailContainerHookView';
export * from './baseReviewDialogCommentHookView';
export * from './baseReviewDialogHookView';
