{% extends "actions/menu_action.html" %}
{% load avatars %}


{% block action_menu_label %}
{%  if siteconfig_settings.avatars_enabled %}
{%   avatar user 32 %}
{%  endif %}
{{request.user.username}}
{% endblock action_menu_label %}
