"use strict";

$(document).ready(function () {
  $('.user').user_infobox();
  $('.bug').bug_infobox();
  $('.review-request-link').review_request_infobox();
  $('time.timesince').timesince();
});

//# sourceMappingURL=init.js.map