{% load static %}
<li class="rb-c-sidebar__nav-item{% if current %} -is-active{% endif %}">
{%  if icon %}
 <img src="{% static icon %}" alt="{{name}}" />
{%  endif %}
 <a class="rb-c-sidebar__item-label" href="{{url}}">{{name}}</a>
</li>
