{% load djblets_utils i18n static %}
{% url 'root' as root_url %}
{% trans 'Home' as home_text %}

{% definevar "logo_png" %}{% static "rb/images/logo.png" %}{% enddefinevar %}

<div class="rb-c-topbar__product-info" id="rbinfo">
 <a href="{{root_url}}" aria-label="{{home_text}}">
  <img class="rb-c-topbar__product-logo"
       id="logo"
       aria-hidden="true"
       src="{{logo_png}}"
       srcset="{{logo_png}} 1x, {% static "rb/images/logo@2x.png" %} 2x"
       alt=""
       border="0"
       width="60"
       height="57">
 </a>
 <h1 class="rb-c-topbar__product-name" id="title">
  <a href="{{root_url}}" aria-label="{{home_text}}">
   {{PRODUCT_NAME}}
   <span class="rb-c-topbar__product-suffix">
    {{settings.PRODUCT_SUFFIX}}
   </span>
  </a>
  <span class="rb-c-topbar__product-version">
   {{version}}
   {% if section_title %} - <strong>{{section_title}}</strong>{% endif %}
  </span>
 </h1>
</div>
