/****************************************************************************
 * Security checklist
 ****************************************************************************/

.admin-security-checklist {
  margin: 0;
  padding: 0;

  .admin-security-check {
    list-style: none;
    margin-bottom: 4em;
    line-height: 140%;

    &:last-child {
      margin-bottom: 0;
    }

    h3 {
      margin-top: 0;
      margin-bottom: 0;
    }

    .admin-security-check-status {
      font-weight: bold;

      &.pass {
        color: var(--ink-p-accent-success-fg-on-container);
      }

      &.fail {
        color: var(--ink-p-accent-error-fg-on-container);
      }
    }

    .admin-security-check-result {
      margin: 1em 0 0 0em;
    }
  }
}
