/**
 * A model backing a widget's state and operations.
 *
 * This stores basic state information on a widget. Subclasses can override
 * this to provide additional state or operations for interacting with the
 * widget or its data on the server.
 */
RB.Admin.Widget = Backbone.Model.extend();
