{% extends "ui/components/page-banner.html" %}
{% load i18n %}


{% block banner_id %}read-only-banner{% endblock %}


{% block banner_header %}
{%  blocktrans %}
{{PRODUCT_NAME}} is in read-only mode.
{%  endblocktrans %}
{% endblock banner_header %}


{% block banner_content %}
{%  if read_only_message %}
<span>{{read_only_message}}</span>
{%  endif %}
{% endblock banner_content %}
